import tensorflow as tf
from typing import Dict, Callable, Tuple
import logging

from .searcher import Searcher


class BeamSearch(Searcher):
    def __init__(self,
                 beam_width: int,
                 end_index: int,
                 max_decoding_steps: int) -> None:
        self.logger = logging.getLogger(__name__)
        self.beam_width = beam_width
        self._end_index = end_index
        self._max_decoding_steps = max_decoding_steps

    def debug(self, name, value):   
        self.logger.debug(f"Debug {name} {value}")

    def _first_token(self,
                     start_predictions: tf.Tensor,
                     state: Dict[str, tf.Tensor],
                     step: Callable[
                         [tf.Tensor, Dict[str, tf.Tensor]],
                         Tuple[tf.Tensor, Dict[str, tf.Tensor]]]
                     ) -> Tuple[tf.Tensor, tf.Tensor]:
        batch_size = start_predictions.shape[0]
        # shape: (batch_size*beam_width, max_decoding_steps)
        predictions = tf.fill(
            (batch_size*self.beam_width, self._max_decoding_steps),
            tf.constant(self._end_index, tf.int32), name="predictions")
        # shape: (batch_size*beam_width, )
        log_probs = tf.fill(
            (batch_size*self.beam_width, ),
            tf.constant(1, tf.float32), name="log_probs")
        # shape: (batch_size, target_vocab_size+source_seq_len)
        predicted_proba, state = step(start_predictions, state)
        # shape: (batch_size, beam_width) both
        topkpreds, topkidx = tf.math.top_k(
            predicted_proba, self.beam_width)
        # shape: (batch_size*beam_width, )
        topkidx = tf.reshape(
            tf.cast(topkidx, tf.int32), (batch_size*self.beam_width, ))
        # shape: (batch_size*beam_width, )
        topkpreds = tf.reshape(
            topkpreds, (batch_size*self.beam_width, ))
        indices = tf.expand_dims(
            tf.range(batch_size*self.beam_width, dtype=tf.int32), 1)
        # shape: (batch_size*beam_width, 2)
        indices = tf.concat([
            indices,
            tf.fill(
                (batch_size*self.beam_width, 1),
                tf.constant(0, dtype=tf.int32))], 1)
        predictions = tf.tensor_scatter_nd_update(
            predictions, indices, topkidx)
        log_probs = log_probs*topkpreds
        # tf.py_function(self.debug, ['first', predictions[:9]], [])
        for key, state_tensor in state.items():
            state_tensor = tf.repeat(state_tensor, self.beam_width, axis=0)
            state[key] = state_tensor

        return state, predictions, log_probs

    def search(self,
               start_predictions: tf.Tensor,
               start_state: Dict[str, tf.Tensor],
               step: Callable[
                   [tf.Tensor, Dict[str, tf.Tensor]],
                   Tuple[tf.Tensor, Dict[str, tf.Tensor]]]
               ) -> Tuple[tf.Tensor, tf.Tensor]:

        batch_size = start_predictions.shape[0]
        # shape: (batch_size*beam_width, max_decoding_steps),
        #        (batch_size*beam_width, )
        state, predictions, log_probabilities = self._first_token(
            start_predictions, start_state, step)

        for timestep in tf.range(1, self._max_decoding_steps):
            # shape: (batch_size*beam_width, )
            last_predictions = predictions[:, timestep-1]
            # shape: (batch_size*beam_width, target_vocab_size+source_seq_len)
            predicted_proba, state = step(last_predictions, state)
            # shape: (batch_size*beam_width, beam_width) both
            topkpreds, topkidx = tf.math.top_k(
                predicted_proba, self.beam_width)
            # shape: (batch_size, beam_width*beam_width)
            topkidx = tf.reshape(
                tf.cast(topkidx, tf.int32),
                (batch_size, self.beam_width*self.beam_width))
            # shape: (batch_size, beam_width*beam_width)
            topkpreds = tf.reshape(
                topkpreds, (batch_size, self.beam_width*self.beam_width))
            # shape: (batch_size, beam_width*beam_width)
            step_log_probs = tf.reshape(
                tf.repeat(log_probabilities, self.beam_width, axis=0),
                (batch_size, self.beam_width*self.beam_width))
            step_log_probs = step_log_probs + topkpreds

            # shape: (batch_size, beam_width) both
            topsteplog, topsteplogidx = tf.math.top_k(
                step_log_probs, self.beam_width)

            # shape: (batch_size*beam_width, )
            topsteplog = tf.reshape(topsteplog, (batch_size*self.beam_width,))
            log_probabilities = topsteplog

            # shape: (batch_size, beam_width)
            topkidx = tf.gather(topkidx, topsteplogidx, axis=-1, batch_dims=1)
            # shape: (batch_size*beam_width, )
            topkidx = tf.reshape(topkidx, (batch_size*self.beam_width,))
            indices = tf.expand_dims(
                tf.range(batch_size*self.beam_width, dtype=tf.int32), 1)
            # shape: (batch_size*beam_width, 2)
            indices = tf.concat([
                indices,
                tf.fill(
                    (batch_size*self.beam_width, 1),
                    tf.cast(timestep, tf.int32))], 1)
            predictions = tf.tensor_scatter_nd_update(
                predictions, indices, topkidx)

            # tf.py_function(self.debug, ['current', predictions[:9]], [])
            for key, state_tensor in state.items():
                _, *last_dims = state_tensor.shape
                state_tensor = tf.reshape(tf.repeat(
                    state_tensor, self.beam_width, axis=0),
                    (batch_size, self.beam_width*self.beam_width, *last_dims))
                state_tensor = tf.gather(
                    state_tensor, topsteplogidx, axis=1, batch_dims=1)
                state_tensor = tf.reshape(
                    state_tensor, (batch_size*self.beam_width, *last_dims))
                state[key] = state_tensor

            for t in tf.range(timestep):
                prev_preds = predictions[:, t]
                # shape: (batch_size, beam_width*beam_width)
                prev_preds = tf.reshape(tf.repeat(
                    prev_preds, self.beam_width, axis=0),
                    (batch_size, self.beam_width*self.beam_width))
                # shape: (batch_size, beam_width)
                prev_preds = tf.gather(
                    prev_preds, topsteplogidx, axis=-1, batch_dims=1)
                # shape: (batch_size*beam_width, )
                prev_preds = tf.reshape(
                    prev_preds, (batch_size*self.beam_width,))
                indices = tf.expand_dims(
                    tf.range(batch_size*self.beam_width, dtype=tf.int32), 1)
                # shape: (batch_size*beam_width, 2)
                indices = tf.concat([
                    indices,
                    tf.fill(
                        (batch_size*self.beam_width, 1),
                        tf.cast(t, tf.int32))], 1)
                predictions = tf.tensor_scatter_nd_update(
                    predictions, indices, prev_preds)
                # tf.py_function(self.debug, ['backfill', predictions[:9]], [])

        # shape: (batch_size, beam_width, max_decoding_steps)
        predictions = tf.reshape(
            predictions,
            (batch_size, self.beam_width, self._max_decoding_steps))
        # shape: (batch_size, beam_width)
        log_probabilities = tf.reshape(
            log_probabilities, (batch_size, self.beam_width))

        return predictions, log_probabilities



from tensorflow.keras.layers import Embedding


class FixedEmbedding(Embedding):
    def __init__(self, embedding_matrix, max_seq_len=None, **kwargs):
        super().__init__(
            embedding_matrix.shape[0],
            embedding_matrix.shape[1],
            trainable=False,
            input_length=max_seq_len,
            **kwargs
        )
        self.embeddings = embedding_matrix

    def build(self, input_shape):
        # with tf.device('cpu:0'):
        #     self.embeddings = self.add_weight(
        #         shape=(self.input_dim, self.output_dim),
        #         initializer=self.embeddings_initializer,
        #         name='embeddings',
        #         regularizer=self.embeddings_regularizer,
        #         constraint=self.embeddings_constraint,
        #         caching_device='cpu:0')
        self.built = True



from abc import ABC, abstractmethod
import tensorflow as tf
from typing import Callable, Dict


class Searcher(ABC):

    @abstractmethod
    def search(self,
               start_predictions: tf.Tensor,
               decode_step: Callable[[tf.Tensor, Dict[str, tf.Tensor]],
                                     Dict[str, tf.Tensor]],
               state: Dict[str, tf.Tensor]):
        pass



from tensorflow.keras.layers import Dense


class FixedDense(Dense):
    def __init__(self, units, weights, **kwargs):
        super().__init__(
            units,
            trainable=False,
            **kwargs
        )
        self.kernel = weights[0]
        self.bias = weights[1]

    def build(self, input_shape):
        self.built = True



from .fixed_embedding import FixedEmbedding
from .additive_attention import BahdanauAttention
from .fixed_dense import FixedDense


__all__ = [
    "FixedEmbedding",
    "BahdanauAttention",
    "FixedDense",
]


# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Python implementation of BLEU and smooth-BLEU.
This module provides a Python implementation of BLEU and smooth-BLEU.
Smooth BLEU is computed following the method outlined in the paper:
Chin-Yew Lin, Franz Josef Och. ORANGE: a method for evaluating automatic
evaluation metrics for machine translation. COLING 2004.
"""

import collections
import math


def _get_ngrams(segment, max_order):
    """Extracts all n-grams upto a given maximum order from an input segment.
    Args:
        segment: text segment from which n-grams will be extracted.
        max_order: maximum length in tokens of the n-grams returned by this
            methods.
    Returns:
        The Counter containing all n-grams upto max_order in segment
        with a count of how many times each n-gram occurred.
    """
    ngram_counts = collections.Counter()
    for order in range(1, max_order + 1):
        for i in range(0, len(segment) - order + 1):
            ngram = tuple(segment[i:i+order])
            ngram_counts[ngram] += 1
    return ngram_counts


def _preprocess(
        reference_corpus, translation_corpus,
        ignore_tokens, ignore_all_tokens_after):
    if ignore_all_tokens_after is not None:
        reference_corpus_n = []
        for references in reference_corpus:
            references_n = []
            for reference in references:
                i = 0
                while (i < len(reference)
                       and reference[i] != ignore_all_tokens_after):
                    i += 1
                i += 1
                references_n.append(reference[:i])
            reference_corpus_n.append(references_n)
        reference_corpus = reference_corpus_n

        translation_corpus_n = []
        for translation in translation_corpus:
            i = 0
            while (i < len(translation)
                   and translation[i] != ignore_all_tokens_after):
                i += 1
            i += 1
            translation_corpus_n.append(translation[:i])
        translation_corpus = translation_corpus_n

    if ignore_tokens is not None:
        reference_corpus_n = []
        for references in reference_corpus:
            references_n = []
            for reference in references:
                ref = []
                for token in reference:
                    if token not in ignore_tokens:
                        ref.append(token)
                references_n.append(ref)
            reference_corpus_n.append(references_n)
        reference_corpus = reference_corpus_n

        translation_corpus_n = []
        for translation in translation_corpus:
            trans = []
            for token in translation:
                if token not in ignore_tokens:
                    trans.append(token)
            translation_corpus_n.append(trans)
        translation_corpus = translation_corpus_n
    return reference_corpus, translation_corpus


def compute_bleu(reference_corpus, translation_corpus, max_order=4,
                 smooth=False, ignore_tokens=None,
                 ignore_all_tokens_after=None):
    """Computes BLEU score of translated segments against one or more
    references.
    Args:
        reference_corpus: list of lists of references for each translation.
            Each reference should be tokenized into a list of tokens.
        translation_corpus: list of translations to score. Each translation
            should be tokenized into a list of tokens.
        max_order: Maximum n-gram order to use when computing BLEU score.
        smooth: Whether or not to apply Lin et al. 2004 smoothing.
    Returns:
        3-Tuple with the BLEU score, n-gram precisions, geometric mean of
        n-gram precisions and brevity penalty.
    """
    reference_corpus, translation_corpus = _preprocess(
        reference_corpus, translation_corpus,
        ignore_tokens, ignore_all_tokens_after)
    matches_by_order = [0] * max_order
    possible_matches_by_order = [0] * max_order
    reference_length = 0
    translation_length = 0
    for (references, translation) in zip(reference_corpus,
                                         translation_corpus):
        reference_length += min(len(r) for r in references)
        translation_length += len(translation)

        merged_ref_ngram_counts = collections.Counter()
        for reference in references:
            merged_ref_ngram_counts |= _get_ngrams(reference, max_order)
            translation_ngram_counts = _get_ngrams(translation, max_order)
            overlap = translation_ngram_counts & merged_ref_ngram_counts
        for ngram in overlap:
            matches_by_order[len(ngram)-1] += overlap[ngram]
        for order in range(1, max_order+1):
            possible_matches = len(translation) - order + 1
            if possible_matches > 0:
                possible_matches_by_order[order-1] += possible_matches

    precisions = [0] * max_order
    for i in range(0, max_order):
        if smooth:
            precisions[i] = ((matches_by_order[i] + 1.) /
                             (possible_matches_by_order[i] + 1.))
        else:
            if possible_matches_by_order[i] > 0:
                precisions[i] = (float(matches_by_order[i]) /
                                 possible_matches_by_order[i])
            else:
                precisions[i] = 0.0

    if min(precisions) > 0:
        p_log_sum = sum((1. / max_order) * math.log(p) for p in precisions)
        geo_mean = math.exp(p_log_sum)
    else:
        geo_mean = 0

    ratio = float(translation_length) / reference_length

    if ratio > 1.0:
        bp = 1.
    else:
        bp = math.exp(1 - 1. / ratio)

    bleu = geo_mean * bp

    return (bleu, precisions, bp, ratio, translation_length, reference_length)



from .beam_search import BeamSearch


class GreedySearch(BeamSearch):
    def __init__(self, end_index, max_decoding_steps):
        super(GreedySearch, self).__init__(1, end_index, max_decoding_steps)



from collections import defaultdict
import numpy as np
import pickle
from typing import Dict


class Vocab:
    def __init__(self,
                 start_token: str,
                 end_token: str,
                 pad_token: str,
                 unk_token: str,
                 source_seq_len: int,
                 target_seq_len: int,
                 pretrained_vectors: Dict[str, np.ndarray] = None) -> None:
        self._start_token = start_token
        self._end_token = end_token
        self._pad_token = pad_token
        self._unk_token = unk_token
        self._source_seq_len = source_seq_len
        self._target_seq_len = target_seq_len
        if pretrained_vectors is not None:
            self._fit_pretrained(pretrained_vectors)

    def _fit_pretrained(self,
                        pretrained_vectors: Dict[str, np.ndarray]) -> None:
        self._source = {}
        self._source[self._pad_token] = 0
        self._source[self._unk_token] = 1
        self._source[self._start_token] = 2
        self._source[self._end_token] = 3
        i = 4
        for token, vector in pretrained_vectors.items():
            if token in self._source:
                continue
            self._source[token] = i
            i += 1
        self._target = self._source.copy()
        self._pretrained_vectors = pretrained_vectors
        self._source_inverse = {}
        for token, idx in self._source.items():
            self._source_inverse[idx] = token
        self._target_inverse = {}
        for token, idx in self._target.items():
            self._target_inverse[idx] = token

    def fit(self,
            tokenized_source,
            tokenized_target,
            pretrained_vectors,
            min_source_freq=1,
            min_target_freq=5):
        source_vocab = defaultdict(lambda: 0)
        for tokens in tokenized_source:
            for token in tokens:
                source_vocab[token.text] += 1

        target_vocab = defaultdict(lambda: 0)
        for tokens in tokenized_target:
            for token in tokens:
                target_vocab[token.text] += 1

        source_vocab = sorted(source_vocab.items(), key=lambda x: -x[1])
        print(f"source max {source_vocab[0]} min {source_vocab[-1]}")
        if min_source_freq is not None:
            source_vocab = filter(
                lambda x: x[1] > min_source_freq, source_vocab)

        target_vocab = sorted(target_vocab.items(), key=lambda x: -x[1])
        print(f"target max {target_vocab[0]} min {target_vocab[-1]}")
        if min_target_freq is not None:
            target_vocab = filter(
                lambda x: x[1] > min_target_freq, target_vocab)

        self._source = {}
        self._source[self._pad_token] = 0
        self._source[self._unk_token] = 1
        self._source[self._start_token] = 2
        self._source[self._end_token] = 3
        for i, k in enumerate(source_vocab, start=4):
            self._source[k[0]] = i

        self._source_inverse = {}
        for token, idx in self._source.items():
            self._source_inverse[idx] = token

        self._target = {}
        self._target[self._pad_token] = 0
        self._target[self._unk_token] = 1
        self._target[self._start_token] = 2
        self._target[self._end_token] = 3
        for i, k in enumerate(target_vocab, start=4):
            self._target[k[0]] = i

        self._target_inverse = {}
        for token, idx in self._target.items():
            self._target_inverse[idx] = token

        self._pretrained_vectors = pretrained_vectors

    def _transform(self, tokenized, token2index, seq_len):
        res = np.zeros((len(tokenized), seq_len), dtype=np.int64)
        for i, tokens in enumerate(tokenized):
            res[i, 0] = token2index[self._start_token]
            for j, token in enumerate(tokens, start=1):
                if j > seq_len - 2:
                    j -= 1
                    break
                res[i, j] = token2index.get(
                    token.text, token2index[self._unk_token])
            res[i, j+1] = token2index[self._end_token]
        return res

    def transform(self, tokenized, namespace):
        if namespace == 'source':
            return self._transform(
                tokenized, self._source, self._source_seq_len)
        elif namespace == 'target':
            return self._transform(
                tokenized, self._target, self._target_seq_len)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def _get_token_id(self, token, token2index):
        return token2index.get(token, token2index[self._unk_token])

    def get_token_id(self, token, namespace):
        if namespace == 'source':
            return self._get_token_id(token, self._source)
        elif namespace == 'target':
            return self._get_token_id(token, self._target)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def _add_token(self, token, token2index):
        if token in token2index:
            return token2index[token]
        i = len(token2index)
        token2index[token] = i
        return i

    def add_token(self, token, namespace):
        if namespace == 'source':
            return self._get_token_id(token, self._source)
        elif namespace == 'target':
            return self._get_token_id(token, self._target)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def get_sequence_len(self, namespace):
        if namespace == 'source':
            return self._source_seq_len
        elif namespace == 'target':
            return self._target_seq_len
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def _get_token_text(self, idx, idx2token):
        return idx2token.get(idx)

    def get_token_text(self, idx, namespace):
        if namespace == 'source':
            return self._get_token_text(idx, self._source_inverse)
        elif namespace == 'target':
            return self._get_token_text(idx, self._target_inverse)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def _inverse_transform(self, idxed, idx2token):
        res = np.full_like(idxed, 'UNK', dtype=object)
        for i, idxs in enumerate(idxed):
            for j, idx in enumerate(idxs):
                res[i, j] = self._get_token_text(idx, idx2token)
        return res

    def inverse_transform(self, idxed, namespace):
        if namespace == 'source':
            return self._inverse_transform(idxed, self._source_inverse)
        elif namespace == 'target':
            return self._inverse_transform(idxed, self._target_inverse)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def save(self, filename_prefix):
        source_filename = filename_prefix+"_source.pkl"
        target_filename = filename_prefix+"_target.pkl"
        pretrained_vecs_filename = filename_prefix+"_pretrained_vecs.pkl"
        with open(source_filename, "wb") as f:
            pickle.dump(self._source, f, protocol=4)
        with open(target_filename, "wb") as f:
            pickle.dump(self._target, f, protocol=4)
        with open(pretrained_vecs_filename, "wb") as f:
            pickle.dump(self._pretrained_vectors, f, protocol=4)

    @classmethod
    def load(cls,
             start_token,
             end_token,
             pad_token,
             unk_token,
             source_seq_len,
             target_seq_len,
             filename_prefix):
        source_filename = filename_prefix+"_source.pkl"
        target_filename = filename_prefix+"_target.pkl"
        pretrained_vecs_filename = filename_prefix+"_pretrained_vecs.pkl"
        inst = cls(start_token, end_token, pad_token, unk_token,
                   source_seq_len, target_seq_len)
        with open(source_filename, "rb") as f:
            inst._source = pickle.load(f)
        with open(target_filename, "rb") as f:
            inst._target = pickle.load(f)
        with open(pretrained_vecs_filename, "rb") as f:
            inst._pretrained_vectors = pickle.load(f)
        inst._source_inverse = {}
        for token, idx in inst._source.items():
            inst._source_inverse[idx] = token
        inst._target_inverse = {}
        for token, idx in inst._target.items():
            inst._target_inverse[idx] = token
        return inst

    def get_vocab_size(self, namespace):
        if namespace == 'source':
            return len(self._source)
        elif namespace == 'target':
            return len(self._target)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")

    def _get_embedding_matrix(self, idx2token):
        embedding_matrix = np.zeros((len(idx2token), 300), dtype=np.float32)
        for idx, token in idx2token.items():
            try:
                embedding_matrix[idx] = self._pretrained_vectors[token]
            except:  # noqa
                pass
        return embedding_matrix

    def get_embedding_matrix(self, namespace):
        if namespace == 'source':
            return self._get_embedding_matrix(self._source_inverse)
        elif namespace == 'target':
            return self._get_embedding_matrix(self._target_inverse)
        else:
            raise ValueError(f"Unknown namespace: {namespace}")



from tensorflow.keras.layers import Layer
import tensorflow as tf


class BahdanauAttention(Layer):
    def __init__(self, units):
        super(BahdanauAttention, self).__init__()
        self.W1 = tf.keras.layers.Dense(units)
        self.W2 = tf.keras.layers.Dense(units)
        self.V = tf.keras.layers.Dense(1)

    def _verify_shapes(self, query, values, values_mask):
        qbatch_size, qhidden_size = query.shape
        vbatch_size, vseq_len, vhidden_size = values.shape
        vmbatch_size, vmseq_len = (
            values_mask.shape if values_mask is not None else values.shape[:2])
        if not qbatch_size == vbatch_size or not vbatch_size == vmbatch_size:
            raise ValueError("Batch sizes should be same everywhere")
        if qhidden_size != vhidden_size:
            raise ValueError("Hidden dim of query and values should match")
        if vseq_len != vmseq_len:
            raise ValueError(
                "Sequence length of values and values_mask should match")

    def call(self, query, values, values_mask=None):
        self._verify_shapes(query, values, values_mask)
        # query hidden state shape == (batch_size, hidden size)
        # query_with_time_axis shape == (batch_size, 1, hidden size)
        # values shape == (batch_size, max_len, hidden size)
        # we are doing this to broadcast addition along the time axis to
        # calculate the score
        query_with_time_axis = tf.expand_dims(query, 1)

        # score shape == (batch_size, max_length, 1)
        # we get 1 at the last axis because we are applying score to self.V
        # the shape of the tensor before applying self.V is
        # (batch_size, max_length, units)
        score = self.V(tf.nn.tanh(
            self.W1(query_with_time_axis) + self.W2(values)))

        # Make mask scores values highly neg
        if values_mask is not None:
            values_mask = tf.expand_dims(tf.logical_not(values_mask), -1)
            score -= 1.e9 * tf.cast(values_mask, dtype=score.dtype)

        # attention_weights shape == (batch_size, max_length, 1)
        attention_weights = tf.nn.softmax(score, axis=1)

        # context_vector shape after sum == (batch_size, hidden_size)
        context_vector = attention_weights * values
        context_vector = tf.reduce_sum(context_vector, axis=1)

        return context_vector, attention_weights



import tensorflow as tf


def masked_log_softmax(logits: tf.Tensor,
                       mask: tf.Tensor,
                       axis: int = None,
                       name: str = None) -> tf.Tensor:
    if mask is not None:
        # make masked logits highly negative
        logits = logits + tf.math.log(tf.cast(mask, tf.float32) + 1e-35)
    return tf.nn.log_softmax(logits, axis, name)


def masked_softmax(logits: tf.Tensor,
                   mask: tf.Tensor,
                   axis: int = None,
                   name: str = None) -> tf.Tensor:
    if mask is None:
        return tf.nn.softmax(logits, axis, name)
    else:
        # nullify contribution of masked logits
        mask = tf.cast(mask, tf.float32)
        inp = logits*mask
        result = tf.nn.softmax(inp, axis, name)
        result = result * mask
        # ensures addition to 1
        result /= (tf.reduce_sum(result, axis=axis, keepdims=True)+1e-35)
        return result



import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import Dense, GRU
from typing import Dict, Any, Tuple
import logging

from .layers import FixedEmbedding, BahdanauAttention
from .vocab import Vocab
from .search import Searcher
from .util import masked_log_softmax, masked_softmax


class Decoder(Model):
    def __init__(self,
                 vocab: Vocab,
                 encoder_out_dim: int,
                 searcher: Searcher,
                 output_layer: Layer,
                 copy_token: str = "@COPY@",
                 **kwargs: Dict[str, Any]) -> None:
        super(Decoder, self).__init__(**kwargs)
        self.vocab = vocab
        self.logger = logging.getLogger(__name__)
        self._copy_index = self.vocab.add_token(copy_token, "target")
        self._unk_index = self.vocab.get_token_id(
            self.vocab._unk_token, "target")
        self._start_index = self.vocab.get_token_id(
            self.vocab._start_token, "target")
        self._end_index = self.vocab.get_token_id(
            self.vocab._end_token, "target")
        self._pad_index = self.vocab.get_token_id(
            self.vocab._pad_token, "target")
        self._srcunk_index = self.vocab.get_token_id(
            self.vocab._unk_token, "source")
        self._srcstart_index = self.vocab.get_token_id(
            self.vocab._start_token, "source")
        self._srcend_index = self.vocab.get_token_id(
            self.vocab._end_token, "source")
        self._srcpad_index = self.vocab.get_token_id(
            self.vocab._pad_token, "source")
        self._searcher = searcher
        self.encoder_out_dim = encoder_out_dim
        self.decoder_out_dim = self.encoder_out_dim
        self.decoder_inp_dim = self.decoder_out_dim
        self._build_decoder_layers(output_layer)

    def _build_decoder_layers(self, output_layer):
        embedding_matrix = self.vocab.get_embedding_matrix("target")
        self.target_embedder = FixedEmbedding(
            embedding_matrix, self.vocab.get_sequence_len("target"))
        self.decode_attn = BahdanauAttention(self.decoder_out_dim)
        self.decoder_input_projector = Dense(self.decoder_inp_dim)
        self.gru = GRU(self.decoder_out_dim)
        self._target_vocab_size = self.vocab.get_vocab_size("target")
        self.output_generation_layer = output_layer
        self.output_copying_layer = Dense(
            self.decoder_out_dim, activation="tanh")

    def call(self,
             source_token_ids: tf.Tensor,
             source2target_ids: tf.Tensor,
             source_embeddings: tf.Tensor,
             source_mask: tf.Tensor,
             state: Dict[str, tf.Tensor],
             target_token_ids: tf.Tensor = None,
             target2source_ids: tf.Tensor = None,
             training: bool = None) -> Dict[str, tf.Tensor]:
        """
        # Parameters

        source_token_ids : `tf.Tensor`
            Shape: (batch_size, source_seq_len)
        source2target_ids : `tf.Tensor`
            Shape: (batch_size, source_seq_len)
        target_token_ids : `tf.Tensor`
            Shape: (batch_size, target_seq_len)
        target2source_ids : `tf.Tensor`
            Shape: (batch_size, target_seq_len)
        training : `bool`
        """
        state["source_mask"] = source_mask
        state["source_token_ids"] = source_token_ids
        state["source2target_ids"] = source2target_ids
        state["source_embeddings"] = source_embeddings

        if target_token_ids is not None:
            state = self._init_decoder_state(state)
            output_dict = self._calculate_loss(
                target_token_ids, target2source_ids, state)
            # fill dummy values to make `call` and `predict` graph give
            # consistent outputs
            output_dict["predictions"] = 1
            output_dict["predicted_probas"] = 1.0
        else:
            output_dict = {}

        return output_dict

    def predict(self,
                source_token_ids: tf.Tensor,
                source2target_ids: tf.Tensor,
                source_embeddings: tf.Tensor,
                source_mask: tf.Tensor,
                state: Dict[str, tf.Tensor],
                target_token_ids: tf.Tensor = None,
                target2source_ids: tf.Tensor = None,
                training: bool = None) -> Dict[str, tf.Tensor]:
        output_dict = {}
        state = self._init_decoder_state(state)
        output_dict.update(self._decode_output(state))
        # fill dummy values to make `call` and `predict` graph give
        # consistent outputs
        output_dict["loss"] = 1.0
        return output_dict

    def _init_decoder_state(
            self, state: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        # shape: (batch, decoder_out_dim)
        state["decoder_hidden"] = state["encoder_final_output"]
        return state

    def _decode_output(
            self, state: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        batch_size, source_length = state["source_mask"].shape
        # shape: (batch_size, source_length)
        state["copy_log_probs"] = tf.math.log(
            (tf.zeros((batch_size, source_length), dtype=tf.float32) + 1e-35))
        start_predictions = tf.fill(
            (batch_size,), tf.constant(self._start_index, dtype=tf.int32))

        # shape: (batch_size, beam_width, max_decoding_steps)
        # shape: (batch_size, beam_width)
        all_top_k_predictions, log_probabilities = self._searcher.search(
            start_predictions, state, self._take_search_step)

        return {
            "predicted_probas": log_probabilities,
            "predictions": all_top_k_predictions,
        }

    def _take_search_step(self,
                          last_predictions: tf.Tensor,
                          state: Dict[str, tf.Tensor]) -> Tuple[
                              tf.Tensor, Dict[str, tf.Tensor]]:
        _, source_seq_len = state["source_mask"].shape

        # Get input to the decoder RNN and the selective weights.
        # `input_choices` is the result of replacing target OOV tokens in
        # `last_predictions` with the copy symbol. `selective_weights` consist
        # of the normalized copy probabilities assigned to the source tokens
        # that were copied. If no tokens were copied, there will be all zeros.
        # shape: (batch_size,), (batch_size, source_seq_len)
        input_choices, selective_weights = (
            self._get_input_and_selective_weights(last_predictions, state))

        # Update the decoder state by taking a step through the RNN.
        state = self._decoder_step(input_choices, selective_weights, state)
        # Get the un-normalized generation scores for each token in the target
        # vocab.
        # shape: (batch_size, target_vocab_size)
        generation_scores = self._get_generation_scores(state)
        # Get the un-normalized copy scores for each token in the source
        # sentence, excluding the start and end tokens.
        # shape: (batch_size, source_seq_len)
        copy_scores = self._get_copy_score(state)
        # Concat un-normalized generation and copy scores.
        # shape: (batch_size, target_vocab_size + source_seq_len)
        all_scores = tf.concat((generation_scores, copy_scores), -1)
        # shape: (batch_size, source_seq_len)
        copy_mask = tf.cast(state["source_mask"], tf.float32)
        # shape: (batch_size, target_vocab_size + source_seq_len)
        mask = tf.concat(
            [
                tf.fill(
                    generation_scores.shape, tf.constant(1, dtype=tf.float32)),
                copy_mask,
            ],
            axis=-1)
        # Normalize generation and copy scores.
        # shape: (batch_size, target_vocab_size + source_seq_len)
        log_probs = masked_log_softmax(all_scores, mask, axis=1)
        # shape: (batch_size, target_vocab_size), (batch_size, source_seq_len)
        generation_log_probs, copy_log_probs = tf.split(
            log_probs, [self._target_vocab_size, source_seq_len], axis=-1)
        # Update copy_probs needed for getting the `selective_weights` at the
        # next timestep.
        state["copy_log_probs"] = copy_log_probs

        # We now have normalized generation and copy scores, but to produce the
        # final score for each token in the extended vocab, we have to go
        # through and add the copy scores to the generation scores of matching
        # target tokens, and sum the copy scores of duplicate source tokens.
        # shape: (batch_size, target_vocab_size + source_seq_len)
        final_log_probs = self._gather_final_log_probs(
            generation_log_probs, copy_log_probs, state
        )

        return final_log_probs, state

    def _get_input_and_selective_weights(
        self, last_predictions: tf.Tensor, state: Dict[str, tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:

        batch_size, source_seq_len = state["source_mask"].shape

        # This is a mask indicating which last predictions were copied from the
        # the source AND not in the target vocabulary (OOV).
        # (batch_size,)
        only_copied_mask = tf.cast(
            last_predictions >= self._target_vocab_size, tf.int32)

        # If the last prediction was in the target vocab or OOV but not copied,
        # we use that as input, otherwise we use the COPY token.
        # shape: (batch_size,)
        copy_input_choices = tf.fill(
            (batch_size,), value=tf.constant(self._copy_index, dtype=tf.int32))

        input_choices = (
            last_predictions * (1 - only_copied_mask)
            + copy_input_choices * only_copied_mask
        )

        # In order to get the `selective_weights`, we need to find out which
        # predictions were copied or copied AND generated, which is the case
        # when a prediction appears in both the source sentence and the target
        # vocab. But whenever a prediction is in the target vocab (even if it
        # also appeared in the source sentence), its index will be the
        # corresponding target vocab index, not its index in the source
        # sentence offset by the target vocab size. So we first use
        # `state["source2target_ids"]` to get an indicator of every source
        # token that matches the predicted target token.
        # shape: (batch_size, source_seq_len)
        expanded_last_predictions = tf.repeat(
            tf.expand_dims(last_predictions, -1), source_seq_len, axis=-1)
        # shape: (batch_size, source_seq_len)
        source_copied_and_generated = (
            state["source2target_ids"] == expanded_last_predictions
        )

        # In order to get indicators for copied source tokens that are OOV with
        # respect to the target vocab, we'll make use of
        # `state["source_token_ids"]`. First we adjust predictions relative to
        # the start of the source tokens. This makes sense because predictions
        # for copied tokens are given by the index of the copied token in the
        # source sentence, offset by the size of the target vocabulary.
        # shape: (batch_size,)
        adjusted_predictions = last_predictions - self._target_vocab_size
        # The adjusted indices for items that were not copied will be negative
        # numbers, and therefore invalid. So we zero them out.
        adjusted_predictions = adjusted_predictions * only_copied_mask
        # shape: (batch_size, source_seq_len)
        source_token_ids = state["source_token_ids"]
        # shape: (batch_size, 1)
        adjusted_prediction_ids = tf.gather(
            source_token_ids,
            tf.expand_dims(adjusted_predictions, -1), axis=-1, batch_dims=1)
        # This mask will contain indicators for source tokens that were copied
        # during the last timestep.
        # shape: (batch_size, source_seq_len)
        source_only_copied = source_token_ids == adjusted_prediction_ids
        # Since we zero'd-out indices for predictions that were not copied,
        # we need to zero out all entries of this mask corresponding to those
        # predictions.
        source_only_copied = (
            source_only_copied
            & tf.cast(tf.expand_dims(only_copied_mask, -1), bool))

        # shape: (batch_size, source_seq_len)
        mask = source_only_copied | source_copied_and_generated
        # shape: (batch_size, source_seq_len)
        selective_weights = masked_softmax(
            state["copy_log_probs"], mask, axis=1)

        return input_choices, selective_weights

    def _gather_final_log_probs(
        self,
        generation_log_probs: tf.Tensor,
        copy_log_probs: tf.Tensor,
        state: Dict[str, tf.Tensor],
    ) -> tf.Tensor:
        _, source_seq_len = state["source_mask"].shape
        source_token_ids = state["source_token_ids"]

        # shape: [(batch_size, *)]
        modified_log_probs_list = []
        for i in range(source_seq_len):
            # shape: (batch_size,)
            copy_log_probs_slice = copy_log_probs[:, i]
            # `source2target_ids` is a matrix of shape (batch_size,
            # source_seq_len) where element (i, j) is the vocab index of the
            # target token that matches the jth source token in the ith group,
            # if there is one, or the index of the OOV symbol otherwise. We'll
            # use this to add copy scores to corresponding generation scores.
            # shape: (batch_size,)
            source2target_ids_slice = state["source2target_ids"][:, i]
            # The OOV index in the source2target_ids_slice indicates that the
            # source token is not in the target vocab, so we don't want to add
            # that copy score to the OOV token.
            copy_log_probs_to_add_mask = tf.cast(
                source2target_ids_slice != self._unk_index, tf.float32)
            copy_log_probs_to_add = (
                copy_log_probs_slice
                + tf.math.log(copy_log_probs_to_add_mask + 1e-35)
            )
            # shape: (batch_size, 1)
            copy_log_probs_to_add = tf.expand_dims(copy_log_probs_to_add, -1)
            # shape: (batch_size, 1)
            selected_generation_log_probs = tf.gather(
                generation_log_probs,
                tf.expand_dims(source2target_ids_slice, -1),
                axis=1, batch_dims=1)
            combined_scores = tf.math.reduce_logsumexp(
                tf.concat([
                    selected_generation_log_probs, copy_log_probs_to_add],
                    axis=1),
                axis=-1
            )
            indices = tf.expand_dims(tf.range(
                source2target_ids_slice.shape[0],
                dtype=source2target_ids_slice.dtype), -1)
            indices = tf.concat(
                [indices, tf.expand_dims(source2target_ids_slice, -1)], -1)
            generation_log_probs = tf.tensor_scatter_nd_update(
                generation_log_probs, indices, combined_scores)
            # We have to combine copy scores for duplicate source tokens so
            # that we can find the overall most likely source token. So, if
            # this is the first occurence of this particular source token, we
            # add the log_probs from all other occurences, otherwise we zero it
            # out since it was already accounted for.
            if i < (source_seq_len - 1):
                # Sum copy scores from future occurences of source token.
                # shape: (batch_size, source_seq_len - i)
                source_future_occurences = tf.cast(
                    source_token_ids[:, (i + 1):]
                    == tf.expand_dims(source_token_ids[:, i], -1), tf.float32)
                # shape: (batch_size, source_seq_len - i)
                future_copy_log_probs = (
                    copy_log_probs[:, (i + 1):]
                    + tf.math.log(source_future_occurences + 1e-35)
                )
                # shape: (batch_size, 1 + source_seq_len - i)
                combined = tf.concat(
                    [tf.expand_dims(copy_log_probs_slice, -1),
                     future_copy_log_probs], axis=-1
                )
                # shape: (batch_size,)
                copy_log_probs_slice = tf.math.reduce_logsumexp(
                    combined, axis=-1)
            if i > 0:
                # Remove copy log_probs that we have already accounted for.
                # shape: (batch_size, i)
                source_previous_occurences = tf.cast(source_token_ids[
                    :, 0:i
                ] == tf.expand_dims(source_token_ids[:, i], -1), tf.int32)
                # shape: (batch_size,)
                duplicate_mask = tf.cast(tf.reduce_sum(
                    source_previous_occurences, axis=-1) == 0, tf.float32)
                copy_log_probs_slice = (
                    copy_log_probs_slice + tf.math.log(duplicate_mask + 1e-35)
                )

            # Finally, we zero-out copy scores that we added to the generation
            # scores above so that we don't double-count them.
            # shape: (batch_size,)
            left_over_copy_log_probs = (
                copy_log_probs_slice
                + tf.math.log(1 - copy_log_probs_to_add_mask + 1e-35)
            )
            modified_log_probs_list.append(
                tf.expand_dims(left_over_copy_log_probs, -1))
        modified_log_probs_list.insert(0, generation_log_probs)

        # shape: (batch_size, target_vocab_size + source_seq_len)
        modified_log_probs = tf.concat(modified_log_probs_list, axis=-1)

        return modified_log_probs

    def _calculate_loss(self,
                        target_token_ids: tf.Tensor,
                        target2source_ids: tf.Tensor,
                        state: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        """
        # Parameters

        target_token_ids : `tf.Tensor`
            Shape: (batch_size, target_seq_len)
        target2source_ids : `tf.Tensor`
            Shape: (batch_size, target_seq_len)
        state : `Dict[str, tf.Tensor]`
        """
        batch_size, target_seq_len = target_token_ids.shape
        num_decoding_steps = target_seq_len - 1
        # shape: (batch, source_seq_len)
        source2target_slice = tf.zeros_like(
            state["source_token_ids"], dtype=tf.float32)

        # mask 1 if source token is not unknown
        # shape: (batch, source_seq_len)
        source_unk_mask = ~(state["source_token_ids"] == self._unk_index)

        # shape: (batch,)
        copy_slice = tf.fill((batch_size, ), tf.constant(
            self._copy_index, dtype=tf.int32))

        # shape: (batch, source_seq_len)
        selective_weights = tf.zeros_like(
            state["source_mask"], dtype=tf.float32)

        # shape: (batch, source_seq_len)
        copy_mask = tf.cast(state["source_mask"], tf.float32)

        # shape: (batch, target_vocab_size)
        generation_mask = tf.ones(
            (batch_size, self._target_vocab_size), dtype=tf.float32)

        # shape: (batch, num_decoding_steps)
        step_log_likelihoods = tf.zeros(
            (batch_size, num_decoding_steps), dtype=tf.float32)
        for timestep in tf.range(num_decoding_steps):
            # shape: (batch,)
            target_input_slice = target_token_ids[:, timestep]

            if timestep < num_decoding_steps - 1:
                # shape: (batch,)
                copied = tf.cast((
                    (target_input_slice == self._unk_index)
                    & (tf.reduce_sum(source2target_slice, -1) > 0)
                ), tf.int32)

                target_input_slice = (
                    target_input_slice * (1 - copied) + copy_slice * copied
                )

                # shape: (batch, source_seq_len)
                source2target_slice = (
                    state["source_token_ids"]
                    == tf.expand_dims(target2source_ids[:, timestep + 1], 1))

                # shape: (batch, source_seq_len)
                source2target_slice = source2target_slice & source_unk_mask

                # shape: (batch, source_seq_len)
                source2target_slice = tf.cast(source2target_slice, tf.float32)

            # Update decoder hidden state by stepping through the decoder
            state = self._decoder_step(
                target_input_slice, selective_weights, state)

            # shape: (batch, target_vocab_size)
            generation_scores = self._get_generation_scores(state)

            # shape: (batch, source_seq_len)
            copy_scores = self._get_copy_score(state)

            # shape: (batch,)
            target_output_slice = target_token_ids[:, timestep+1]

            # shape: (batch,), (batch, source_seq_len)
            step_log_likelihood, selective_weights = self._get_ll_contrib(
                generation_scores, copy_scores, generation_mask,
                copy_mask, target_output_slice, source2target_slice)

            # shape: (batch, 1)
            step_log_likelihood = tf.expand_dims(step_log_likelihood, 1)

            if timestep > 0:
                step_log_likelihoods = tf.concat(
                    [step_log_likelihoods[:, :timestep],
                     step_log_likelihood,
                     step_log_likelihoods[:, timestep+1:]], 1)
                step_log_likelihoods.set_shape(
                    (batch_size, num_decoding_steps))
            else:
                step_log_likelihoods = tf.concat(
                    [step_log_likelihood,
                     step_log_likelihoods[:, 1:]], 1)
            # step_log_likelihoods = tf.concat(
            #     [step_log_likelihoods, step_log_likelihood], 1)

        # shape: (batch, num_decoding_steps)
        log_likelihoods = step_log_likelihoods

        # shape: (batch, target_seq_len)
        target_mask = tf.cast(target_token_ids != self._pad_index, tf.float32)

        # shape: (batch, num_decoding_steps)
        target_mask = target_mask[:, 1:]

        # shape: (batch,)
        log_likelihood = tf.reduce_sum(log_likelihoods*target_mask, axis=-1)

        # The loss is the negative log-likelihood, averaged over the batch.
        # shape: ()
        loss = -tf.reduce_sum(log_likelihood) / batch_size

        return {"loss": loss}

    def _decoder_step(self,
                      last_predictions: tf.Tensor,
                      selective_weights: tf.Tensor,
                      state: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        """
        # Parameters

        last_predictions : `tf.Tensor`
            Shape : (batch,)
        selective_weights : `tf.Tensor`
            Shape : (batch, source_seq_len)
        state : `Dict[str, tf.Tensor]`
        """
        # (batch, source_seq_len)
        source_mask = state["source_mask"]

        # shape: (batch, 1)
        last_predictions = tf.expand_dims(last_predictions, 1)
        # shape: (batch, 1, target_emb_dim)
        target_embedded = self.target_embedder(last_predictions)
        # shape: (batch, decoder_out_dim)
        query = state["decoder_hidden"]
        # shape: (batch, decoder_out_dim), (batch, source_seq_len, 1)
        attentive_read, attentive_weights = self.decode_attn(
            query, state["encoder_output"], source_mask)

        # shape: (batch, 1, decoder_out_dim)
        attentive_read = tf.expand_dims(attentive_read, 1)

        # shape: (batch, 1, source_seq_len)
        selective_weights = tf.expand_dims(selective_weights, -2)
        # shape: (batch, 1, encoder_out_dim)
        selective_read = tf.matmul(selective_weights, state["encoder_output"])

        # shape: (batch, 1, target_emb_dim+encoder_out_dim+decoder_out_dim)
        decoder_input = tf.concat([
            target_embedded, attentive_read, selective_read], -1)
        projected_decoder_input = self.decoder_input_projector(decoder_input)

        # shape: (batch, decoder_out_dim)
        state["decoder_hidden"] = self.gru(
            projected_decoder_input, initial_state=state["decoder_hidden"])
        return state

    def _get_generation_scores(
            self, state: Dict[str, tf.Tensor]) -> tf.Tensor:
        return self.output_generation_layer(state["decoder_hidden"])

    def _get_copy_score(self, state: Dict[str, tf.Tensor]) -> tf.Tensor:
        # shape: (batch, source_seq_len, encoder_out_dim)
        encoder_output = state["encoder_output"]
        # shape: (batch, source_seq_len, decoder_out_dim)
        copy_projection = self.output_copying_layer(encoder_output)

        # shape: (batch, decoder_out_dim, 1)
        decoder_hidden = tf.expand_dims(state["decoder_hidden"], -1)

        # shape: (batch, source_seq_len, 1)
        copy_scores = tf.matmul(copy_projection, decoder_hidden)

        # shape: (batch, source_seq_len)
        copy_scores = tf.squeeze(copy_scores, axis=-1)

        return copy_scores

    def debug(self, name, value):
        self.logger.debug(f"Debug {name} {value}")

    def _get_ll_contrib(self,
                        generation_scores: tf.Tensor,
                        copy_scores: tf.Tensor,
                        generation_mask: tf.Tensor,
                        copy_mask: tf.Tensor,
                        target_token_id: tf.Tensor,
                        source2target: tf.Tensor) -> Tuple[
                            tf.Tensor, tf.Tensor]:
        _, target_size = generation_scores.shape
        # tf.py_function(
        #     self.debug, ['target_token_id', target_token_id[:3]], [])
        # tf.py_function(
        #     self.debug, ['source2target', source2target[:3]], [])

        # shape: (batch, target_vocab_size + source_seq_len)
        mask = tf.concat([generation_mask, copy_mask], -1)

        # shape: (batch, target_vocab_size + source_seq_len)
        all_scores = tf.concat([generation_scores, copy_scores], -1)

        # Normalize generation and copy scores.
        # shape: (batch_size, target_vocab_size + source_seq_len)
        log_probs = masked_log_softmax(all_scores, mask, axis=1)

        # Calculate the log probability (`copy_log_probs`) for each token in
        # the source sentence that matches the current target token. We use the
        # sum of these copy probabilities for matching tokens in the source
        # sentence to get the total probability for the target token. We also
        # need to normalize the individual copy probabilities to create
        # `selective_weights`, which are used in the next timestep to create
        # a selective read state.
        # shape: (batch, source_seq_len)
        copy_log_probs = (
            log_probs[:, target_size:] + tf.math.log(
                tf.cast(source2target, tf.float32) + 1e-35)
        )
        # Since `log_probs[:, target_size]` gives us the raw copy log
        # probabilities, we use a non-log softmax to get the normalized non-log
        # copy probabilities.
        selective_weights = masked_softmax(
            log_probs[:, target_size:], source2target, axis=1)
        # tf.py_function(
        #     self.debug, ['selective_weights', selective_weights[:3]], [])
        # This mask ensures that item in the batch has a non-zero generation
        # probabilities for this timestep only when the real target token is
        # not OOV or there are no matching tokens in the source sentence.
        # shape: (batch_size,)
        gen_mask = tf.cast(
            (target_token_id != self._unk_index)
            | (tf.reduce_sum(source2target, -1) == 0), tf.float32)
        # shape: (batch_size, 1)
        log_gen_mask = tf.expand_dims(tf.math.log(gen_mask + 1e-35), 1)

        # shape: (batch_size, 1)
        generation_log_probs = (
            tf.gather(
                log_probs, tf.expand_dims(target_token_id, 1), axis=1,
                batch_dims=1)
            + log_gen_mask
        )
        # shape: (batch, 1+source_seq_len)
        combined_gen_copy = tf.concat(
            [generation_log_probs, copy_log_probs], -1)
        # tf.py_function(
        #     self.debug, ['combined_gen_copy', combined_gen_copy[:3]], [])

        # shape: (batch,)
        step_log_likelihood = tf.math.reduce_logsumexp(
            combined_gen_copy, axis=-1)

        # tf.py_function(
        #     self.debug, ['step_log_likelihood', step_log_likelihood[:3]], [])

        return step_log_likelihood, selective_weights



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
par_directory = path.abspath(path.dirname(__file__))
with open(path.join(par_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="copynet-tf",
    description="CopyNet with TensorFlow 2.0",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/pavanchhatpar/copynet-tf",
    packages=find_packages(),
    version="0.1.0",
    author='Pavan Chhatpar',
    author_email='pavanchhatpar@gmail.com'
)
